/* wdbinfo v1.0 - extract audio CD track names etc. from a workman database.
 * Copyright (C) 1996 Russell Marks.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 * You don't actually have to have a CD-ROM drive (or to play audio CDs
 * with it) to use this program, but it's difficult to see what use
 * it could be if you don't. :-)
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>

#define WDBINFO_VER		"1.0"


/* XXX want enums really */
#define LIST_ALL		1
#define LIST_CD			2
#define LIST_TRACK		3
#define LIST_ALL_BY_ARTIST	4

#define LIST_FIRST_MATCH	1
#define LIST_ALL_MATCHES	2

#define WHITESPACE_CHARS	" \t"

#define UNKNOWN_STR		"<unknown>"


/* a bit of a global-variable-fest, I'm afraid... */

char infile[1024];
FILE *in;
char buf[256];

/* info on each track on a CD, derived from the DB.
 * dynamically allocated would be nice, but as there's an in-built
 * max. of 255 tracks on a CD, why bother? :-)
 * NB: 0th element is unused.
 */
struct trkinfo_tag
  {
  char name[256];
  int addr;		/* distance in frames from start of CD */
  int len;		/* length in frames */
  } trkinfo[256];

char cd_name[256];
char cd_artist[256];
int cd_len;		/* length of CD in sec */
int ntrks;		/* number of tracks on CD */

/* these altered by cmdline options */
int mode=LIST_CD;
int submode=LIST_FIRST_MATCH;
char *lookup_arg=NULL;
int exact_matching=0;
int verbose=1;


/* use WORKMANDB or ~/.workmandb as default database */
void make_default_infile_name()
{
char *pref=getenv("WORKMANDB");
char *home=getenv("HOME")?getenv("HOME"):"";

if(pref)
  strcpy(infile,pref);
else
  {
  strcpy(infile,home);
  strcat(infile,"/.workmandb");
  }
}


void open_db()
{
if((in=fopen(infile,"rb"))==NULL)
  {
  fprintf(stderr,"Couldn't open database '%s'.\n",infile);
  exit(1);
  }
}


void close_db()
{
fclose(in);
}


void usage_help()
{
printf("Wdbinfo v%s - (c) 1996 Russell Marks for improbabledesigns.\n",
	WDBINFO_VER);
printf(
"
usage: wdbinfo [-aAhlmtx] [-f db_name] [name]

	-a	list all matches to 'name' rather than just the first.
        -A	list all CDs by artists matching 'name'.
        -f	specify workman DB to use instead of default.
	-h	this usage help.
	-l	list all CDs.
        -m	minimal output. When listing CDs, list only name/artist;
        	when listing tracks, list only the track name. Useful
                for shell scripts, sorting, etc.
        -t	list track matching 'name' rather than CD.
        -x	('exact') disable partial matching.

Unless '-l' is specified, you must specify 'name', which if after '-t'
should be a track name, if after '-A' should be an artist name, and
otherwise should be a CD name. It need not be the whole name, any
partial match is sufficient, and case is disregarded. '-a' only applies
to CD/track name matching functions.
");
}


void parse_options(int argc,char **argv)
{
int done=0;

done=0;
opterr=0;

do
  switch(getopt(argc,argv,"aAf:hlmtx"))
    {
    case 'a':	/* list all matches */
      submode=LIST_ALL_MATCHES;
      break;
    case 'A':	/* list all by an artist */
      mode=LIST_ALL_BY_ARTIST;
      break;
    case 'f':	/* specify different DB to use */
      strcpy(infile,optarg);
      break;
    case 'h':
      usage_help();
      exit(1);
    case 'l':	/* list all CD names and contents */
      mode=LIST_ALL;
      break;
    case 'm':	/* not verbose :-) */
      verbose=0;
      break;
    case 't':
      mode=LIST_TRACK;
      break;
    case 'x':
      exact_matching=1;
      break;
    case '?':
      switch(optopt)
        {
        case 'f':
          fprintf(stderr,"The -f option must be followed by a filename.\n");
          break;
        default:
          fprintf(stderr,"Option '%c' not recognised.\n",optopt);
        }
      exit(1);
    case -1:
      done=1;
    }
while(!done);

/* if no filename given, and one was needed... */
if(mode!=LIST_ALL)
  {
  if(optind!=argc-1)
    {
    fprintf(stderr,"Need a%s name! (try 'wdbinfo -h' for help)\n",
    	(mode==LIST_TRACK)?" track":(mode==LIST_CD)?" CD":"n artist");
    exit(1);
    }
  else
    lookup_arg=argv[optind];
  }
}


int getline()
{
char *ptr;

buf[0]=0;
do
  ptr=fgets(buf,256,in);
while(buf[0]=='#' && !feof(in));

if(ptr!=NULL)
  {
  if(buf[0]!=0 && buf[strlen(buf)-1]=='\n') buf[strlen(buf)-1]=0;
  }
else
  buf[0]=0;
return((ptr==NULL)?0:1);
}


int get_tracks_line()
{
while(getline()!=0)
  if(strncmp(buf,"tracks ",7)==0)
    return(1);

return(0);
}


/* move to first whitespace character.
 * returns NULL if end of string found or bad 'pos'.
 */
char *next_end_of_word(char *pos)
{
if(pos==NULL) return(pos);
while(*pos && strchr(WHITESPACE_CHARS,*pos)==NULL) pos++;
return((*pos)?pos:NULL);
}


/* move to first non-whitespace (letter, number, etc.) character.
 * returns NULL if end of string found or bad 'pos'.
 */
char *next_start_of_word(char *pos)
{
if(pos==NULL) return(pos);
while(*pos && strchr(WHITESPACE_CHARS,*pos)!=NULL) pos++;
return((*pos)?pos:NULL);
}


/* move to next word (assuming 'pos' points to the start of a word). */
char *next_word(char *pos)
{
return(next_start_of_word(next_end_of_word(pos)));
}


/* fill in the various bits possible from the tracks line -
 * i.e. num. tracks, track addrs and lengths, and cd length.
 * the tracks line must be in buf before calling.
 * this also blanks out name/artist/trackname in case they're missing
 * (they shouldn't be, but... just in case...).
 */
void parse_tracks_line()
{
char *ptr=buf+7;	/* 7=strlen("tracks ") */
int f;

ntrks=atoi(ptr);

for(f=1;f<=ntrks;f++)
  {
  ptr=next_word(ptr);
  trkinfo[f].addr=atoi(ptr);
  }

ptr=next_word(ptr);
cd_len=atoi(ptr);

/* now fill in track lengths */
for(f=1;f<=ntrks;f++)
  if(f<ntrks)
    {
    /* for all but last track, use offset in frames of next track
     * to get length of this one.
     */
    trkinfo[f].len=trkinfo[f+1].addr-trkinfo[f].addr;
    }
  else
    {
    /* for the last track, we have to use the CD length. This is
     * only accurate to the nearest second.
     */
    trkinfo[f].len=cd_len*75-trkinfo[f].addr;
    }

/* finally, blank out names */
strcpy(cd_name,UNKNOWN_STR);
strcpy(cd_artist,UNKNOWN_STR);
for(f=1;f<=ntrks;f++) strcpy(trkinfo[f].name,UNKNOWN_STR);
}


/* look ahead for 'cdname' or 'tracks' or EOF.
 * it then resets to old file pos.
 */
void set_cd_name()
{
long oldpos=ftell(in);

while(getline()!=0 && strncmp(buf,"tracks ",7)!=0)
  {
  if(strncmp(buf,"cdname ",7)==0)
    {
    strcpy(cd_name,buf+7);
    fseek(in,oldpos,SEEK_SET);
    return;
    }
  }

/* if it gets here, we didn't find one; just leave it as 'unknown'. */
fseek(in,oldpos,SEEK_SET);
}


/* look ahead for 'artist' or 'tracks' or EOF.
 * it then resets to old file pos.
 */
void set_cd_artist()
{
long oldpos=ftell(in);

while(getline()!=0 && strncmp(buf,"tracks ",7)!=0)
  {
  if(strncmp(buf,"artist ",7)==0)
    {
    strcpy(cd_artist,buf+7);
    fseek(in,oldpos,SEEK_SET);
    return;
    }
  }

/* if it gets here, we didn't find one; just leave it as 'unknown'. */
fseek(in,oldpos,SEEK_SET);
}


/* look ahead for all 'track' tokens up until 'tracks' or EOF.
 * this doesn't reset old file pos, since it's the last thing to be read.
 */
void set_cd_tracknames()
{
int trk=1;
long oldpos=ftell(in);

while(getline()!=0 && strncmp(buf,"tracks ",7)!=0)
  {
  if(strncmp(buf,"track ",6)==0)
    {
    strcpy(trkinfo[trk].name,buf+6);
    trk++;
    }
  }

fseek(in,oldpos,SEEK_SET);
}


/* read all the info on a CD from the DB.
 * use after a successful call to get_tracks_line().
 */
void read_cd_info()
{
parse_tracks_line();
set_cd_name();
set_cd_artist();
set_cd_tracknames();
}


void smash_case(char *str)
{
int f;

for(f=0;f<strlen(str);f++)
  str[f]=tolower(str[f]);
}


/* check cd/track name against lookup_arg, ignoring case. */
int check_name(char *name)
{
static char larg[256];
static int first_time=1;
char tbuf[256];

if(first_time)
  {
  /* fill in larg[] with case-smashed lookup_arg */
  strcpy(larg,lookup_arg);
  smash_case(larg);
  first_time=0;
  }

strcpy(tbuf,name);
smash_case(tbuf);
if(exact_matching)
  {
  if(strcmp(tbuf,larg)==0) return(1);
  }
else
  if(strstr(tbuf,larg)) return(1);

return(0);
}


void show_cd()
{
int f,tmp;

if(verbose)
  {
  printf("-- \"%s\" by %s --\n",cd_name,cd_artist);
  
  /* I prefer to stick to min:sec even for more than an hour */
  printf("Total play time: %2d:%02d\n\n",cd_len/60,cd_len%60);
  
  for(f=1;f<=ntrks;f++)
    {
    tmp=trkinfo[f].len/75;
    printf("%3d. (%2d:%02d) %s\n",f,tmp/60,tmp%60,trkinfo[f].name);
    }
  
  printf("\n");
  }
else
  {
  /* not verbose - tabular format */
  printf("%-23s\t%-55s\n",cd_artist,cd_name);
  }
}


/* this one shows a certain track number along with the album it's from */
void show_track(int trknum)
{
int tmp=trkinfo[trknum].len/75;

if(verbose)
  {
  printf("- (%2d:%02d) %s\n",tmp/60,tmp%60,trkinfo[trknum].name);
  printf("  (from \"%s\" by %s)\n\n",cd_name,cd_artist);
  }
else
  {
  /* not verbose - just the raw track name */
  puts(trkinfo[trknum].name);
  }
}


void list_all()
{
while(get_tracks_line())
  {
  read_cd_info();
  show_cd();
  }
}


void find_cd()
{
while(get_tracks_line())
  {
  read_cd_info();
  if(check_name(cd_name))
    {
    show_cd();
    if(submode==LIST_FIRST_MATCH) return;
    }
  }
}


void find_artist()
{
while(get_tracks_line())
  {
  read_cd_info();
  if(check_name(cd_artist))
    show_cd();
  }
}


void find_track()
{
int f;

while(get_tracks_line())
  {
  read_cd_info();
  for(f=1;f<=ntrks;f++)
    if(check_name(trkinfo[f].name))
      {
      show_track(f);
      if(submode==LIST_FIRST_MATCH) return;
      }
  }
}


int main(int argc,char *argv[])
{
make_default_infile_name();
parse_options(argc,argv);
open_db();

switch(mode)
  {
  case LIST_ALL:
    list_all();
    break;
  case LIST_ALL_BY_ARTIST:
    find_artist();
    break;
  case LIST_CD:
    find_cd();
    break;
  case LIST_TRACK:
    find_track();
    break;
  default:
    fprintf(stderr,"Can't happen.\n");
    exit(1);
  }

close_db();
exit(0);
}
